@@ -4,11 +4,11 @@
 /* SUMMARY
 /*	Postfix queue control
 /* SYNOPSIS
-/*	\fBpostqueue\fR [\fB-v\fR] [\fB-c \fIconfig_dir\fR] \fB-f\fR
+/*	\fBpostqueue\fR [\fB-c \fIconfig_dir\fR] \fB-f\fR
 /* .br
-/*	\fBpostqueue\fR [\fB-v\fR] [\fB-c \fIconfig_dir\fR] \fB-p\fR
+/*	\fBpostqueue\fR [\fB-c \fIconfig_dir\fR] \fB-p\fR
 /* .br
-/*	\fBpostqueue\fR [\fB-v\fR] [\fB-c \fIconfig_dir\fR] \fB-s \fIsite\fR
+/*	\fBpostqueue\fR [\fB-c \fIconfig_dir\fR] \fB-s \fIsite\fR
 /* DESCRIPTION
 /*	The \fBpostqueue\fR(1) command implements the Postfix user interface
 /*	for queue management. It implements operations that are
@@ -61,8 +61,7 @@
 /*	command, by contacting the Postfix \fBflush\fR(8) daemon.
 /* .IP \fB-v\fR
 /*	Enable verbose logging for debugging purposes. Multiple \fB-v\fR
-/*	options make the software increasingly verbose. As of Postfix 2.3,
-/*	this option is available for the super-user only.
+/*	options make the software increasingly verbose.
 /* SECURITY
 /* .ad
 /* .fi
@@ -184,13 +183,13 @@
 #include <mail_params.h>
 #include <mail_conf.h>
 #include <mail_task.h>
+#include <debug_process.h>
 #include <mail_run.h>
 #include <mail_flush.h>
 #include <flush_clnt.h>
 #include <smtp_stream.h>
 #include <user_acl.h>
 #include <valid_mailhost_addr.h>
-#include <mail_dict.h>
 
 /* Application-specific. */
 
@@ -363,13 +362,6 @@ static void flush_site(const char *site)
     }
 }
 
-/* unavailable - sanitize exit status from library run-time errors */
-
-static void unavailable(void)
-{
-    exit(EX_UNAVAILABLE);
-}
-
 /* usage - scream and die */
 
 static NORETURN usage(void)
@@ -388,6 +380,7 @@ int     main(int argc, char **argv)
     int     mode = PQ_MODE_DEFAULT;
     char   *site_to_flush = 0;
     ARGV   *import_env;
+    char   *last;
     int     bad_site;
 
     /*
@@ -413,7 +406,6 @@ int     main(int argc, char **argv)
     if ((slash = strrchr(argv[0], '/')) != 0 && slash[1])
 	argv[0] = slash + 1;
     msg_vstream_init(argv[0], VSTREAM_ERR);
-    msg_cleanup(unavailable);
     msg_syslog_init(mail_task("postqueue"), LOG_PID, LOG_FACILITY);
     set_mail_conf_str(VAR_PROCNAME, var_procname = mystrdup(argv[0]));
 
@@ -446,8 +438,7 @@ int     main(int argc, char **argv)
 	    site_to_flush = optarg;
 	    break;
 	case 'v':
-	    if (geteuid() == 0)
-		msg_verbose++;
+	    msg_verbose++;
 	    break;
 	default:
 	    usage();
@@ -460,7 +451,6 @@ int     main(int argc, char **argv)
      * Further initialization...
      */
     mail_conf_read();
-    mail_dict_init();				/* proxy, sql, ldap */
     get_mail_conf_str_table(str_table);
 
     /*
