@@ -2,33 +2,31 @@
 /* NAME
 /*	smtp_connect 3
 /* SUMMARY
-/*	connect to SMTP/LMTP server and deliver
+/*	connect to SMTP server and deliver
 /* SYNOPSIS
 /*	#include "smtp.h"
 /*
 /*	int	smtp_connect(state)
 /*	SMTP_STATE *state;
 /* DESCRIPTION
-/*	This module implements SMTP/LMTP connection management and controls
+/*	This module implements SMTP connection management and controls
 /*	mail delivery.
 /*
-/*	smtp_connect() attempts to establish an SMTP/LMTP session with a host
+/*	smtp_connect() attempts to establish an SMTP session with a host
 /*	that represents the destination domain, or with an optional fallback
-/*	relay when {the destination cannot be found, or when all the
-/*	destination servers are unavailable}. It skips over IP addresses
-/*	that fail to complete the SMTP/LMTP handshake and tries to find
-/*	an alternate server when an SMTP/LMTP session fails to deliver.
+/*	relay when the destination cannot be found, or when all the
+/*	destination servers are unavailable. It skips over IP addresses
+/*	that fail to complete the SMTP handshake and tries to find
+/*	an alternate server when an SMTP session fails to deliver.
 /*
-/*	This layer also controls what connections are retrieved from
-/*	the connection cache, and what connections are saved to the cache.
+/*	This layer also controls what sessions are retrieved from
+/*	the session cache, and what sessions are saved to the cache.
 /*
 /*	The destination is either a host (or domain) name or a numeric
 /*	address. Symbolic or numeric service port information may be
-/*	appended, separated by a colon (":"). In the case of LMTP,
-/*	destinations may be specified as "unix:pathname", "inet:host"
-/*	or "inet:host:port".
+/*	appended, separated by a colon (":").
 /*
-/*	With SMTP, the Internet domain name service is queried for mail
+/*	By default, the Internet domain name service is queried for mail
 /*	exchanger hosts. Quote the domain name with `[' and `]' to
 /*	suppress mail exchanger lookups.
 /*
@@ -57,7 +55,6 @@
 #include <sys_defs.h>
 #include <stdlib.h>
 #include <sys/socket.h>
-#include <sys/un.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <errno.h>
@@ -94,7 +91,6 @@
 #include <own_inet_addr.h>
 #include <deliver_pass.h>
 #include <mail_error.h>
-#include <dsn_buf.h>
 
 /* DNS library. */
 
@@ -106,80 +102,47 @@
 #include <smtp_addr.h>
 #include <smtp_reuse.h>
 
- /*
-  * Forward declaration.
-  */
-static SMTP_SESSION *smtp_connect_sock(int, struct sockaddr *, int,
-				               const char *, const char *,
-				               unsigned,
-				               const char *, DSN_BUF *,
-				               int);
+#define STR(x) vstring_str(x)
 
-/* smtp_connect_unix - connect to UNIX-domain address */
+/* smtp_salvage - salvage the server reply before disconnecting */
 
-static SMTP_SESSION *smtp_connect_unix(const char *addr,
-				               DSN_BUF *why,
-				               int sess_flags)
+static VSTRING *smtp_salvage(VSTREAM *stream)
 {
-    const char *myname = "smtp_connect_unix";
-    struct sockaddr_un sock_un;
-    int     len = strlen(addr);
-    int     sock;
-
-    dsb_reset(why);				/* Paranoia */
+    int     len = vstream_peek(stream);
+    VSTRING *buf = vstring_alloc(len);
 
     /*
-     * Sanity checks.
+     * We know the server replied with 4... or 5...; salvage whatever we have
+     * received in the VSTREAM buffer and sanitize any non-printable crud.
      */
-    if (len >= (int) sizeof(sock_un.sun_path)) {
-	msg_warn("unix-domain name too long: %s", addr);
-	dsb_simple(why, "4.3.5", "Server configuration error");
-	return (0);
-    }
-
-    /*
-     * Initialize.
-     */
-    memset((char *) &sock_un, 0, sizeof(sock_un));
-    sock_un.sun_family = AF_UNIX;
-#ifdef HAS_SUN_LEN
-    sock_un.sun_len = len + 1;
-#endif
-    memcpy(sock_un.sun_path, addr, len + 1);
-
-    /*
-     * Create a client socket.
-     */
-    if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
-	msg_fatal("%s: socket: %m", myname);
-
-    /*
-     * Connect to the server.
-     */
-    if (msg_verbose)
-	msg_info("%s: trying: %s...", myname, addr);
-
-    return (smtp_connect_sock(sock, (struct sockaddr *) & sock_un,
-			      sizeof(sock_un), var_myhostname, addr,
-			      0, addr, why, sess_flags));
+    vstream_fread(stream, STR(buf), len);
+    VSTRING_AT_OFFSET(buf, len);		/* XXX not public interface */
+    VSTRING_TERMINATE(buf);
+    translit(STR(buf), "\r\n", "  ");
+    printable(STR(buf), '?');
+    return (buf);
 }
 
 /* smtp_connect_addr - connect to explicit address */
 
-static SMTP_SESSION *smtp_connect_addr(const char *destination, DNS_RR *addr,
-				               unsigned port, DSN_BUF *why,
+static SMTP_SESSION *smtp_connect_addr(const char *dest, DNS_RR *addr,
+				               unsigned port, VSTRING *why,
 				               int sess_flags)
 {
-    const char *myname = "smtp_connect_addr";
+    char   *myname = "smtp_connect_addr";
     struct sockaddr_storage ss;		/* remote */
     struct sockaddr *sa = (struct sockaddr *) & ss;
     SOCKADDR_SIZE salen = sizeof(ss);
     MAI_HOSTADDR_STR hostaddr;
     int     sock;
+    int     conn_stat;
+    int     saved_errno;
+    VSTREAM *stream;
+    int     ch;
     char   *bind_addr;
     char   *bind_var;
 
-    dsb_reset(why);				/* Paranoia */
+    smtp_errno = SMTP_ERR_NONE;			/* Paranoia */
 
     /*
      * Sanity checks.
@@ -187,7 +150,7 @@ static SMTP_SESSION *smtp_connect_addr(const char *destination, DNS_RR *addr,
     if (dns_rr_to_sa(addr, port, sa, &salen) != 0) {
 	msg_warn("%s: skip address type %s: %m",
 		 myname, dns_strtype(addr->type));
-	dsb_simple(why, "4.4.0", "network address conversion failed: %m");
+	smtp_errno = SMTP_ERR_RETRY;
 	return (0);
     }
 
@@ -260,33 +223,12 @@ static SMTP_SESSION *smtp_connect_addr(const char *destination, DNS_RR *addr,
     }
 
     /*
-     * Connect to the server.
+     * Connect to the SMTP server.
      */
     SOCKADDR_TO_HOSTADDR(sa, salen, &hostaddr, (MAI_SERVPORT_STR *) 0, 0);
     if (msg_verbose)
 	msg_info("%s: trying: %s[%s] port %d...",
-		 myname, SMTP_HNAME(addr), hostaddr.buf, ntohs(port));
-
-    return (smtp_connect_sock(sock, sa, salen, SMTP_HNAME(addr), hostaddr.buf,
-			      port, destination, why, sess_flags));
-}
-
-/* smtp_connect_sock - connect a socket over some transport */
-
-static SMTP_SESSION *smtp_connect_sock(int sock, struct sockaddr * sa,
-				               int salen, const char *name,
-				               const char *addr,
-				               unsigned port,
-				               const char *destination,
-				               DSN_BUF *why,
-				               int sess_flags)
-{
-    int     conn_stat;
-    int     saved_errno;
-    VSTREAM *stream;
-    time_t  start_time;
-
-    start_time = time((time_t *) 0);
+		 myname, addr->name, hostaddr.buf, ntohs(port));
     if (var_smtp_conn_tmout > 0) {
 	non_blocking(sock, NON_BLOCKING);
 	conn_stat = timed_connect(sock, sa, salen, var_smtp_conn_tmout);
@@ -297,30 +239,59 @@ static SMTP_SESSION *smtp_connect_sock(int sock, struct sockaddr * sa,
 	conn_stat = sane_connect(sock, sa, salen);
     }
     if (conn_stat < 0) {
-	dsb_simple(why, "4.4.1", "connect to %s[%s]: %m", name, addr);
+	vstring_sprintf(why, "connect to %s[%s]: %m",
+			addr->name, hostaddr.buf);
+	smtp_errno = SMTP_ERR_RETRY;
 	close(sock);
 	return (0);
     }
+
+    /*
+     * Skip this host if it takes no action within some time limit.
+     */
+    if (read_wait(sock, var_smtp_helo_tmout) < 0) {
+	vstring_sprintf(why, "connect to %s[%s]: read timeout",
+			addr->name, hostaddr.buf);
+	smtp_errno = SMTP_ERR_RETRY;
+	close(sock);
+	return (0);
+    }
+
+    /*
+     * Skip this host if it disconnects without talking to us.
+     */
     stream = vstream_fdopen(sock, O_RDWR);
+    if ((ch = VSTREAM_GETC(stream)) == VSTREAM_EOF) {
+	vstring_sprintf(why, "connect to %s[%s]: server dropped connection without sending the initial SMTP greeting",
+			addr->name, hostaddr.buf);
+	smtp_errno = SMTP_ERR_RETRY;
+	vstream_fclose(stream);
+	return (0);
+    }
+    vstream_ungetc(stream, ch);
 
     /*
-     * Avoid poor performance when TCP MSS > VSTREAM_BUFSIZE.
+     * Skip this host if it sends a 4xx or 5xx greeting. This prevents us
+     * from counting it towards the MX session limit. Unfortunately, this
+     * also means that we have to salvage the server's response ourself so
+     * that it can be included in logging or in non-delivery reports. It does
+     * not hurt if we keep the test for a 4xx or 5xx greeting in smtp_helo().
      */
-    if (sa->sa_family == AF_INET
-#ifdef AF_INET6
-	|| sa->sa_family == AF_INET6
-#endif
-	)
-	vstream_tweak_tcp(stream);
+    if (ch == '4' || (ch == '5' && var_smtp_skip_5xx_greeting)) {
+	VSTRING *salvage_buf = smtp_salvage(stream);
 
-    /*
-     * Bundle up what we have into a nice SMTP_SESSION object.
-     */
-    return (smtp_session_alloc(stream, destination, name, addr,
-			       port, start_time, sess_flags));
+	vstring_sprintf(why, "connect to %s[%s]: server refused to talk to me: %s",
+			addr->name, hostaddr.buf, STR(salvage_buf));
+	vstring_free(salvage_buf);
+	smtp_errno = SMTP_ERR_RETRY;
+	vstream_fclose(stream);
+	return (0);
+    }
+    return (smtp_session_alloc(stream, dest, addr->name,
+			       hostaddr.buf, port, sess_flags));
 }
 
-/* smtp_parse_destination - parse host/port destination */
+/* smtp_parse_destination - parse destination */
 
 static char *smtp_parse_destination(char *destination, char *def_service,
 				            char **hostp, unsigned *portp)
@@ -340,13 +311,13 @@ static char *smtp_parse_destination(char *destination, char *def_service,
      * destination argument so the parsing can be destructive.
      */
     if ((err = host_port(buf, hostp, (char *) 0, &service, def_service)) != 0)
-	msg_fatal("%s in server description: %s", err, destination);
+	msg_fatal("%s in SMTP server description: %s", err, destination);
 
     /*
      * Convert service to port number, network byte order.
      */
     if (alldig(service)) {
-	if ((port = atoi(service)) >= 65536 || port == 0)
+	if ((port = atoi(service)) >= 65536)
 	    msg_fatal("bad network port in destination: %s", destination);
 	*portp = htons(port);
     } else {
@@ -361,9 +332,7 @@ static char *smtp_parse_destination(char *destination, char *def_service,
 
 static void smtp_cleanup_session(SMTP_STATE *state)
 {
-    DELIVER_REQUEST *request = state->request;
     SMTP_SESSION *session = state->session;
-    int     bad_session;
 
     /*
      * Inform the postmaster of trouble.
@@ -385,144 +354,24 @@ static void smtp_cleanup_session(SMTP_STATE *state)
      * hosts. In fact, this is the only benefit of caching logical to
      * physical bindings; caching a session under its own hostname provides
      * no performance benefit, given the way smtp_connect() works.
+     * 
+     * XXX Should not cache TLS sessions unless we are using a single-session,
+     * in-process, cache. And if we did, we should passivate VSTREAM objects
+     * in addition to passivating SMTP_SESSION objects.
      */
-    bad_session = THIS_SESSION_IS_BAD;		/* smtp_quit() may fail */
-    if (THIS_SESSION_IS_EXPIRED)
-	smtp_quit(state);			/* also disables caching */
-    if (THIS_SESSION_IS_CACHED
-	/* Redundant tests for safety... */
-	&& vstream_ferror(session->stream) == 0
-	&& vstream_ftimeout(session->stream) == 0
-	&& vstream_feof(session->stream) == 0) {
+    if (session->reuse_count > 0) {
 	smtp_save_session(state);
+	if (HAVE_NEXTHOP_STATE(state))
+	    FREE_NEXTHOP_STATE(state);
     } else {
 	smtp_session_free(session);
     }
     state->session = 0;
 
-    /*
-     * If this session was good, reset the logical next-hop state, so that we
-     * won't cache connections to alternate servers under the logical
-     * next-hop destination. Otherwise we could end up skipping over the
-     * available and more preferred servers.
-     */
-    if (HAVE_NEXTHOP_STATE(state) && !bad_session)
-	FREE_NEXTHOP_STATE(state);
-
     /*
      * Clean up the lists with todo and dropped recipients.
      */
     smtp_rcpt_cleanup(state);
-
-    /*
-     * Reset profiling info.
-     * 
-     * XXX When one delivery request results in multiple sessions, the set-up
-     * and transmission latencies of the earlier sessions will count as
-     * connection set-up time for the later sessions.
-     * 
-     * XXX On the other hand, when we first try to connect to one or more dead
-     * hosts before we reach a good host, then all that time must be counted
-     * as connection set-up time for the session with the good host.
-     * 
-     * XXX So this set-up attribution problem exists only when we actually
-     * engage in a session, spend a lot of time delivering a message, find
-     * that it fails, and then connect to an alternate host.
-     */
-    memset((char *) &request->msg_stats.conn_setup_done, 0,
-	   sizeof(request->msg_stats.conn_setup_done));
-    memset((char *) &request->msg_stats.deliver_done, 0,
-	   sizeof(request->msg_stats.deliver_done));
-    request->msg_stats.reuse_count = 0;
-}
-
-/* smtp_connect_local - connect to local server */
-
-static void smtp_connect_local(SMTP_STATE *state, const char *path)
-{
-    const char *myname = "smtp_connect_local";
-    DELIVER_REQUEST *request = state->request;
-    SMTP_SESSION *session;
-    DSN_BUF *why = state->why;
-
-    /*
-     * It's too painful to weave this code into the SMTP connection
-     * management routine.
-     * 
-     * Connection cache management is based on the UNIX-domain pathname, without
-     * the "unix:" prefix.
-     * 
-     * XXX Disable connection caching when sender-dependent authentication is
-     * enabled. We must not send someone elses mail over an authenticated
-     * connection, and we must not send mail that requires authentication
-     * over a connection that wasn't authenticated.
-     */
-#define CAN_ENABLE_CONN_CACHE(request, dest) \
-    (!var_smtp_sender_auth \
-     && ((var_smtp_cache_demand && (request->flags & DEL_REQ_FLAG_SCACHE)) \
-	 || (smtp_cache_dest && string_list_match(smtp_cache_dest, dest))))
-
-    if (CAN_ENABLE_CONN_CACHE(request, path))
-	state->misc_flags |= SMTP_MISC_FLAG_CONN_CACHE;
-
-    /*
-     * XXX We assume that the session->addr member refers to a copy of the
-     * UNIX-domain pathname, so that smtp_save_session() will cache the
-     * connection using the pathname as the physical endpoint name.
-     */
-#define NO_PORT	0
-
-    /*
-     * Opportunistic TLS for unix domain sockets does not make much sense,
-     * since the channel is private, mere encryption without authentication
-     * is just wasted cycles and opportunity for breakage. Since we are not
-     * willing to retry after TLS handshake failures here, we downgrade "may"
-     * no "none". Nothing is lost, and much waste is avoided.
-     * 
-     * We don't know who is authenticating whom, so if a client cert is
-     * available, "encrypt" may be a sensible policy. Otherwise, we also
-     * downgrade "encrypt" to "none", this time just to avoid waste.
-     */
-    if ((state->misc_flags & SMTP_MISC_FLAG_CONN_CACHE) == 0
-	|| (session = smtp_reuse_addr(state, path, NO_PORT)) == 0)
-	session = smtp_connect_unix(path, why, state->misc_flags);
-    if ((state->session = session) != 0) {
-	session->state = state;
-#ifdef USE_TLS
-	session->tls_nexthop = var_myhostname;	/* for TLS_LEV_SECURE */
-	if (session->tls_level == TLS_LEV_MAY) {
-	    msg_warn("%s: opportunistic TLS encryption is not appropriate "
-		     "for unix-domain destinations.", myname);
-	    session->tls_level = TLS_LEV_NONE;
-	}
-#endif
-	/* All delivery errors bounce or defer. */
-	state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;
-
-	/*
-	 * When a TLS handshake fails, the stream is marked "dead" to avoid
-	 * further I/O over a broken channel.
-	 */
-	if ((session->features & SMTP_FEATURE_FROM_CACHE) == 0
-	    && smtp_helo(state) != 0) {
-	    if (!THIS_SESSION_IS_DEAD
-		&& vstream_ferror(session->stream) == 0
-		&& vstream_ftimeout(session->stream) == 0
-		&& vstream_feof(session->stream) == 0)
-		smtp_quit(state);
-	} else {
-	    smtp_xfer(state);
-	}
-
-	/*
-	 * With opportunistic TLS disabled we don't expect to be asked to
-	 * retry connections without TLS, and so we expect the final server
-	 * flag to stay on.
-	 */
-	if ((state->misc_flags & SMTP_MISC_FLAG_FINAL_SERVER) == 0)
-	    msg_panic("%s: unix-domain destination not final!", myname);
-	smtp_cleanup_session(state);
-    }
 }
 
 /* smtp_scrub_address_list - delete all cached addresses from list */
@@ -608,7 +457,7 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
     int     session_count = 0;
     DNS_RR *addr;
     DNS_RR *next;
-    MAI_HOSTADDR_STR hostaddr;
+    int     saved_final_server = state->final_server;
     SMTP_SESSION *session;
 
     /*
@@ -622,9 +471,7 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
     && (session = smtp_reuse_domain(state, lookup_mx, domain, port)) != 0) {
 	session_count = 1;
 	smtp_update_addr_list(addr_list, session->addr, session_count);
-	if ((state->misc_flags & SMTP_MISC_FLAG_FINAL_NEXTHOP)
-	    && *addr_list == 0)
-	    state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;
+	state->final_server = (saved_final_server && *addr_list == 0);
 	smtp_xfer(state);
 	smtp_cleanup_session(state);
     }
@@ -640,16 +487,13 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
 	if (addr->pref != domain_best_pref)
 	    break;
 	next = addr->next;
-	if (dns_rr_to_pa(addr, &hostaddr) != 0
-	    && (session = smtp_reuse_addr(state, hostaddr.buf, port)) != 0) {
+	if ((session = smtp_reuse_addr(state, addr, port)) != 0) {
 	    session->features |= SMTP_FEATURE_BEST_MX;
 	    session_count += 1;
 	    smtp_update_addr_list(addr_list, session->addr, session_count);
 	    if (*addr_list == 0)
 		next = 0;
-	    if ((state->misc_flags & SMTP_MISC_FLAG_FINAL_NEXTHOP)
-		&& next == 0)
-		state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;
+	    state->final_server = (saved_final_server && next == 0);
 	    smtp_xfer(state);
 	    smtp_cleanup_session(state);
 	}
@@ -657,18 +501,31 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
     return (session_count);
 }
 
-/* smtp_connect_remote - establish remote connection */
+/* smtp_connect - establish SMTP connection */
 
-static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
-				        char *def_service)
+int     smtp_connect(SMTP_STATE *state)
 {
     DELIVER_REQUEST *request = state->request;
+    VSTRING *why = vstring_alloc(10);
+    char   *dest_buf;
+    char   *domain;
+    unsigned port;
+    char   *def_service = "smtp";	/* XXX ##IPPORT_SMTP? */
     ARGV   *sites;
     char   *dest;
     char  **cpp;
+    DNS_RR *addr_list;
+    DNS_RR *addr;
+    DNS_RR *next;
+    int     addr_count;
+    int     sess_count;
+    int     misc_flags = SMTP_MISC_FLAG_DEFAULT;
+    SMTP_SESSION *session;
+    int     lookup_mx;
+    unsigned domain_best_pref;
+    int     sess_flags = SMTP_SESS_FLAG_NONE;
+    int     i_am_mx = 0;
     int     non_fallback_sites;
-    int     retry_plain = 0;
-    DSN_BUF *why = state->why;
 
     /*
      * First try to deliver to the indicated destination, then try to deliver
@@ -678,12 +535,11 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
      * primary destination to be a list (it could be just separators).
      */
     sites = argv_alloc(1);
-    argv_add(sites, nexthop, (char *) 0);
+    argv_add(sites, request->nexthop, (char *) 0);
     if (sites->argc == 0)
-	msg_panic("null destination: \"%s\"", nexthop);
+	msg_panic("null destination: \"%s\"", request->nexthop);
     non_fallback_sites = sites->argc;
-    if ((state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) == 0)
-	argv_split_append(sites, var_fallback_relay, ", \t\r\n");
+    argv_split_append(sites, var_fallback_relay, ", \t\r\n");
 
     /*
      * Don't give up after a hard host lookup error until we have tried the
@@ -704,24 +560,10 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 #define IS_FALLBACK_RELAY(cpp, sites, non_fallback_sites) \
 	    (*(cpp) && (cpp) >= (sites)->argv + (non_fallback_sites))
 
-    for (cpp = sites->argv, (state->misc_flags |= SMTP_MISC_FLAG_FIRST_NEXTHOP);
-	 SMTP_RCPT_LEFT(state) > 0 && (dest = *cpp) != 0;
-	 cpp++, (state->misc_flags &= ~SMTP_MISC_FLAG_FIRST_NEXTHOP)) {
-	char   *dest_buf;
-	char   *domain;
-	unsigned port;
-	DNS_RR *addr_list;
-	DNS_RR *addr;
-	DNS_RR *next;
-	int     addr_count;
-	int     sess_count;
-	SMTP_SESSION *session;
-	int     lookup_mx;
-	unsigned domain_best_pref;
-	MAI_HOSTADDR_STR hostaddr;
-
-	if (cpp[1] == 0)
-	    state->misc_flags |= SMTP_MISC_FLAG_FINAL_NEXTHOP;
+    for (cpp = sites->argv; SMTP_RCPT_LEFT(state) > 0 && (dest = *cpp) != 0; cpp++) {
+	if (i_am_mx && IS_FALLBACK_RELAY(cpp, sites, non_fallback_sites))
+	    break;
+	state->final_server = (cpp[1] == 0);
 
 	/*
 	 * Parse the destination. Default is to use the SMTP port. Look up
@@ -736,40 +578,18 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 	 */
 	if (msg_verbose)
 	    msg_info("connecting to %s port %d", domain, ntohs(port));
-	if ((state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) == 0) {
-	    if (ntohs(port) == IPPORT_SMTP)
-		state->misc_flags |= SMTP_MISC_FLAG_LOOP_DETECT;
-	    else
-		state->misc_flags &= ~SMTP_MISC_FLAG_LOOP_DETECT;
-	    lookup_mx = (var_disable_dns == 0 && *dest != '[');
-	} else
-	    lookup_mx = 0;
+	if (ntohs(port) != IPPORT_SMTP)
+	    misc_flags &= ~SMTP_MISC_FLAG_LOOP_DETECT;
+	else
+	    misc_flags |= SMTP_MISC_FLAG_LOOP_DETECT;
+	lookup_mx = (var_disable_dns == 0 && *dest != '[');
 	if (!lookup_mx) {
-	    addr_list = smtp_host_addr(domain, state->misc_flags, why);
+	    addr_list = smtp_host_addr(domain, misc_flags, why);
 	    /* XXX We could be an MX host for this destination... */
 	} else {
-	    int     i_am_mx = 0;
-
-	    addr_list = smtp_domain_addr(domain, state->misc_flags,
-					 why, &i_am_mx);
-	    /* If we're MX host, don't connect to non-MX backups. */
-	    if (i_am_mx)
-		state->misc_flags |= SMTP_MISC_FLAG_FINAL_NEXTHOP;
+	    addr_list = smtp_domain_addr(domain, misc_flags, why, &i_am_mx);
 	}
 
-	/*
-	 * Don't try fall-back hosts if mail loops to myself. That would just
-	 * make the problem worse.
-	 */
-	if (addr_list == 0 && SMTP_HAS_LOOP_DSN(why))
-	    state->misc_flags |= SMTP_MISC_FLAG_FINAL_NEXTHOP;
-
-	/*
-	 * No early loop exit or we have a memory leak with dest_buf.
-	 */
-	if (addr_list)
-	    domain_best_pref = addr_list->pref;
-
 	/*
 	 * When session caching is enabled, store the first good session for
 	 * this delivery request under the next-hop destination name. All
@@ -788,29 +608,40 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 	 * Opportunistic (a.k.a. on-demand) session caching on request by the
 	 * queue manager. This is turned temporarily when a destination has a
 	 * high volume of mail in the active queue.
-	 * 
-	 * XXX Disable connection caching when sender-dependent authentication
-	 * is enabled. We must not send someone elses mail over an
-	 * authenticated connection, and we must not send mail that requires
-	 * authentication over a connection that wasn't authenticated.
 	 */
-	if (addr_list && (state->misc_flags & SMTP_MISC_FLAG_FIRST_NEXTHOP)
-	    && CAN_ENABLE_CONN_CACHE(request, domain)) {
-	    state->misc_flags |= SMTP_MISC_FLAG_CONN_CACHE;
+	if (cpp == sites->argv
+	    && ((var_smtp_cache_demand && (request->flags & DEL_REQ_FLAG_SCACHE) != 0)
+		|| (smtp_cache_dest && string_list_match(smtp_cache_dest, domain)))) {
+	    sess_flags |= SMTP_SESS_FLAG_CACHE;
 	    SET_NEXTHOP_STATE(state, lookup_mx, domain, port);
 	}
 
+	/*
+	 * Don't try any backup host if mail loops to myself. That would just
+	 * make the problem worse.
+	 */
+	if (addr_list == 0 && smtp_errno == SMTP_ERR_LOOP) {
+	    myfree(dest_buf);
+	    break;
+	}
+
+	/*
+	 * No early loop exit or we have a memory leak with dest_buf.
+	 */
+	if (addr_list)
+	    domain_best_pref = addr_list->pref;
+
 	/*
 	 * Delete visited cached hosts from the address list.
 	 * 
 	 * Optionally search the connection cache by domain name or by primary
-	 * MX address before we try to create new connections.
+	 * MX address.
 	 * 
 	 * Enforce the MX session and MX address counts per next-hop or
 	 * fall-back destination. smtp_reuse_session() will truncate the
 	 * address list when either limit is reached.
 	 */
-	if (addr_list && state->misc_flags & SMTP_MISC_FLAG_CONN_CACHE) {
+	if (addr_list && (sess_flags & SMTP_SESS_FLAG_CACHE) != 0) {
 	    if (state->cache_used->used > 0)
 		smtp_scrub_addr_list(state->cache_used, &addr_list);
 	    sess_count = addr_count =
@@ -820,8 +651,7 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 	    sess_count = addr_count = 0;
 
 	/*
-	 * Connect to an SMTP server: create primary MX connections, and
-	 * reuse or create backup MX connections.
+	 * Connect to an SMTP server.
 	 * 
 	 * At the start of an SMTP session, all recipients are unmarked. In the
 	 * course of an SMTP session, recipients are marked as KEEP (deliver
@@ -841,108 +671,51 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 	    next = addr->next;
 	    if (++addr_count == var_smtp_mxaddr_limit)
 		next = 0;
-	    if ((state->misc_flags & SMTP_MISC_FLAG_CONN_CACHE) == 0
+	    if ((sess_flags & SMTP_SESS_FLAG_CACHE) == 0
 		|| addr->pref == domain_best_pref
-		|| dns_rr_to_pa(addr, &hostaddr) == 0
-		|| !(session = smtp_reuse_addr(state, hostaddr.buf, port)))
-		session = smtp_connect_addr(dest, addr, port, why,
-					    state->misc_flags);
+		|| (session = smtp_reuse_addr(state, addr, port)) == 0)
+		session = smtp_connect_addr(dest, addr, port, why, sess_flags);
 	    if ((state->session = session) != 0) {
-		session->state = state;
+		if (++sess_count == var_smtp_mxsess_limit)
+		    next = 0;
+		state->final_server = (cpp[1] == 0 && next == 0);
 		if (addr->pref == domain_best_pref)
 		    session->features |= SMTP_FEATURE_BEST_MX;
-		/* Don't count handshake errors towards the session limit. */
-		if ((state->misc_flags & SMTP_MISC_FLAG_FINAL_NEXTHOP)
-		    && next == 0)
-		    state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;
-#ifdef USE_TLS
-		/* Disable TLS when retrying after a handshake failure */
-		if (retry_plain) {
-		    if (session->tls_level >= TLS_LEV_ENCRYPT)
-			msg_panic("Plain-text retry wrong for mandatory TLS");
-		    session->tls_level = TLS_LEV_NONE;
-		    retry_plain = 0;
-		}
-		session->tls_nexthop = domain;	/* for TLS_LEV_SECURE */
-#endif
-		if ((session->features & SMTP_FEATURE_FROM_CACHE) == 0
-		    && smtp_helo(state) != 0) {
-#ifdef USE_TLS
-
-		    /*
-		     * When an opportunistic TLS handshake fails, try the
-		     * same address again, with TLS disabled. See also the
-		     * RETRY_AS_PLAINTEXT macro.
-		     */
-		    if ((retry_plain = session->tls_retry_plain) != 0) {
-			--addr_count;
-			next = addr;
-		    }
-#endif
-
-		    /*
-		     * When a TLS handshake fails, the stream is marked
-		     * "dead" to avoid further I/O over a broken channel.
-		     */
-		    if (!THIS_SESSION_IS_DEAD
-			&& vstream_ferror(session->stream) == 0
-			&& vstream_ftimeout(session->stream) == 0
-			&& vstream_feof(session->stream) == 0)
-			smtp_quit(state);
-		} else {
-		    /* Do count delivery errors towards the session limit. */
-		    if (++sess_count == var_smtp_mxsess_limit)
-			next = 0;
-		    if ((state->misc_flags & SMTP_MISC_FLAG_FINAL_NEXTHOP)
-			&& next == 0)
-			state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;
+		if ((session->features & SMTP_FEATURE_FROM_CACHE) != 0
+		    || smtp_helo(state, misc_flags) == 0)
 		    smtp_xfer(state);
-		}
 		smtp_cleanup_session(state);
 	    } else {
-		msg_info("%s (port %d)", STR(why->reason), ntohs(port));
+		msg_info("%s (port %d)", vstring_str(why), ntohs(port));
 	    }
-	    /* Insert: test if we must skip the remaining MX hosts. */
 	}
 	dns_rr_free(addr_list);
 	myfree(dest_buf);
-	if (state->misc_flags & SMTP_MISC_FLAG_FINAL_NEXTHOP)
-	    break;
     }
 
     /*
      * We still need to deliver, bounce or defer some left-over recipients:
      * either mail loops or some backup mail server was unavailable.
+     * 
+     * Pay attention to what could be configuration problems, and pretend that
+     * these are recoverable rather than bouncing the mail.
      */
     if (SMTP_RCPT_LEFT(state) > 0) {
+	switch (smtp_errno) {
 
-	/*
-	 * In case of a "no error" indication we make up an excuse: we did
-	 * find the host address, but we did not attempt to connect to it.
-	 * This can happen when the fall-back relay was already tried via a
-	 * cached connection, so that the address list scrubber left behind
-	 * an empty list.
-	 */
-	if (!SMTP_HAS_DSN(why)) {
-	    dsb_simple(why, "4.3.0",
-		       "server unavailable or unable to receive mail");
-	}
+	default:
+	    msg_panic("smtp_connect: bad error indication %d", smtp_errno);
 
-	/*
-	 * Pay attention to what could be configuration problems, and pretend
-	 * that these are recoverable rather than bouncing the mail.
-	 */
-	else if (!SMTP_HAS_SOFT_DSN(why)
-		 && (state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) == 0) {
+	case SMTP_ERR_LOOP:
+	case SMTP_ERR_FAIL:
 
 	    /*
 	     * The fall-back destination did not resolve as expected, or it
 	     * is refusing to talk to us, or mail for it loops back to us.
 	     */
 	    if (IS_FALLBACK_RELAY(cpp, sites, non_fallback_sites)) {
-		msg_warn("%s configuration problem", VAR_SMTP_FALLBACK);
-		vstring_strcpy(why->status, "4.3.5");
-		/* XXX Keep the diagnostic code and MTA. */
+		msg_warn("%s configuration problem", VAR_FALLBACK_RELAY);
+		smtp_errno = SMTP_ERR_RETRY;
 	    }
 
 	    /*
@@ -951,21 +724,38 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
 	     */
 	    else if (strcmp(sites->argv[0], var_relayhost) == 0) {
 		msg_warn("%s configuration problem", VAR_RELAYHOST);
-		vstring_strcpy(why->status, "4.3.5");
-		/* XXX Keep the diagnostic code and MTA. */
+		smtp_errno = SMTP_ERR_RETRY;
 	    }
 
 	    /*
 	     * Mail for the next-hop destination loops back to myself. Pass
 	     * the mail to the best_mx_transport or bounce it.
 	     */
-	    else if (SMTP_HAS_LOOP_DSN(why) && *var_bestmx_transp) {
-		dsb_reset(why);			/* XXX */
+	    else if (smtp_errno == SMTP_ERR_LOOP && *var_bestmx_transp) {
 		state->status = deliver_pass_all(MAIL_CLASS_PRIVATE,
 						 var_bestmx_transp,
 						 request);
 		SMTP_RCPT_LEFT(state) = 0;	/* XXX */
+		break;
 	    }
+	    /* FALLTHROUGH */
+
+	case SMTP_ERR_RETRY:
+
+	    /*
+	     * We still need to bounce or defer some left-over recipients:
+	     * either mail loops or some backup mail server was unavailable.
+	     */
+	    state->final_server = 1;		/* XXX */
+	    smtp_site_fail(state, smtp_errno == SMTP_ERR_RETRY ? 450 : 550,
+			   "%s", vstring_str(why));
+
+	    /*
+	     * Sanity check. Don't silently lose recipients.
+	     */
+	    smtp_rcpt_cleanup(state);
+	    if (SMTP_RCPT_LEFT(state) > 0)
+		msg_panic("smtp_connect: left-over recipients");
 	}
     }
 
@@ -975,75 +765,6 @@ static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
     if (HAVE_NEXTHOP_STATE(state))
 	FREE_NEXTHOP_STATE(state);
     argv_free(sites);
-}
-
-/* smtp_connect - establish SMTP connection */
-
-int     smtp_connect(SMTP_STATE *state)
-{
-    DELIVER_REQUEST *request = state->request;
-    char   *destination = request->nexthop;
-
-    /*
-     * All deliveries proceed along the same lines, whether they are over TCP
-     * or UNIX-domain sockets, and whether they use SMTP or LMTP: get a
-     * connection from the cache or create a new connection; deliver mail;
-     * update the connection cache or disconnect.
-     * 
-     * The major differences appear at a higher level: the expansion from
-     * destination to address list, and whether to stop before we reach the
-     * end of that list.
-     */
-#define DEF_LMTP_SERVICE	var_lmtp_tcp_port
-#define DEF_SMTP_SERVICE	"smtp"
-
-    /*
-     * With LMTP we have direct-to-host delivery only. The destination may
-     * have multiple IP addresses.
-     */
-    if (state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) {
-	if (strncmp(destination, "unix:", 5) == 0) {
-	    smtp_connect_local(state, destination + 5);
-	} else {
-	    if (strncmp(destination, "inet:", 5) == 0)
-		destination += 5;
-	    smtp_connect_remote(state, destination, DEF_LMTP_SERVICE);
-	}
-    }
-
-    /*
-     * With SMTP we can have indirection via MX host lookup, as well as an
-     * optional fall-back relayhost that we must avoid when we are MX host.
-     * 
-     * XXX We don't add support for "unix:" or "inet:" prefixes in SMTP
-     * destinations, because that would break compatibility with existing
-     * Postfix configurations that have a host with such a name.
-     */
-    else {
-	smtp_connect_remote(state, destination, DEF_SMTP_SERVICE);
-    }
-
-    /*
-     * We still need to bounce or defer some left-over recipients: either
-     * (SMTP) mail loops or some server was unavailable.
-     * 
-     * We could avoid this (and the "final server" complexity) by keeping one
-     * DSN structure per recipient in memory, by updating those in-memory
-     * structures with each delivery attempt, and by always flushing all
-     * deferred recipients at the end. We'd probably still want to bounce
-     * recipients immediately, so we'd end up with another chunk of code for
-     * defer logging only.
-     */
-    if (SMTP_RCPT_LEFT(state) > 0) {
-	state->misc_flags |= SMTP_MISC_FLAG_FINAL_SERVER;	/* XXX */
-	smtp_sess_fail(state);
-
-	/*
-	 * Sanity check. Don't silently lose recipients.
-	 */
-	smtp_rcpt_cleanup(state);
-	if (SMTP_RCPT_LEFT(state) > 0)
-	    msg_panic("smtp_connect: left-over recipients");
-    }
+    vstring_free(why);
     return (state->status);
 }
